"""
kernels.py
==========

Kernel functions

Created by Maxim Ziatdinov (email: maxim.ziatdinov@ai4microscopy.com)
"""

from typing import Union, Dict, Callable

import math

import jax.numpy as jnp
from jax import jit

kernel_fn_type = Callable[[jnp.ndarray, jnp.ndarray, Dict[str, jnp.ndarray], jnp.ndarray],  jnp.ndarray]


def _sqrt(x, eps=1e-12):
    return jnp.sqrt(x + eps)


def add_jitter(x, jitter=1e-6):
    return x + jitter


def square_scaled_distance(X: jnp.ndarray, Z: jnp.ndarray,
                           lengthscale: Union[jnp.ndarray, float] = 1.
                           ) -> jnp.ndarray:
    """
    Computes a square of scaled distance, :math:`\|\frac{X-Z}{l}\|^2`,
    between X and Z are vectors with :math:`n x num_features` dimensions
    """
    scaled_X = X / lengthscale
    scaled_Z = Z / lengthscale
    X2 = (scaled_X ** 2).sum(1, keepdims=True)
    Z2 = (scaled_Z ** 2).sum(1, keepdims=True)
    XZ = jnp.matmul(scaled_X, scaled_Z.T)
    r2 = X2 - 2 * XZ + Z2.T
    return r2.clip(0)


@jit
def RBFKernel(X: jnp.ndarray, Z: jnp.ndarray,
              params: Dict[str, jnp.ndarray],
              noise: int = 0, **kwargs: float) -> jnp.ndarray:
    """
    Radial basis function kernel

    Args:
        X: 2D vector with *(number of points, number of features)* dimension
        Z: 2D vector with *(number of points, number of features)* dimension
        params: Dictionary with kernel hyperparameters 'k_length' and 'k_scale'
        noise: optional noise vector with dimension (n,)

    Returns:
        Computed kernel matrix betwenen X and Z
    """
    r2 = square_scaled_distance(X, Z, params["k_length"])
    k = params["k_scale"] * jnp.exp(-0.5 * r2)
    if X.shape == Z.shape:
        k += add_jitter(noise, **kwargs) * jnp.eye(X.shape[0])
    return k


@jit
def MaternKernel(X: jnp.ndarray, Z: jnp.ndarray,
                 params: Dict[str, jnp.ndarray],
                 noise: int = 0, **kwargs: float) -> jnp.ndarray:
    """
    Matern52 kernel

    Args:
        X: 2D vector with *(number of points, number of features)* dimension
        Z: 2D vector with *(number of points, number of features)* dimension
        params: Dictionary with kernel hyperparameters 'k_length' and 'k_scale'
        noise: optional noise vector with dimension (n,)

    Returns:
        Computed kernel matrix between X and Z
    """
    r2 = square_scaled_distance(X, Z, params["k_length"])
    r = _sqrt(r2)
    sqrt5_r = 5**0.5 * r
    k = params["k_scale"] * (1 + sqrt5_r + (5/3) * r2) * jnp.exp(-sqrt5_r)
    if X.shape == Z.shape:
        k += add_jitter(noise, **kwargs) * jnp.eye(X.shape[0])
    return k


@jit
def PeriodicKernel(X: jnp.ndarray, Z: jnp.ndarray,
                   params: Dict[str, jnp.ndarray],
                   noise: int = 0, **kwargs: float
                   ) -> jnp.ndarray:
    """
    Periodic kernel

    Args:
        X: 2D vector with *(number of points, number of features)* dimension
        Z: 2D vector with *(number of points, number of features)* dimension
        params: Dictionary with kernel hyperparameters 'k_length', 'k_scale', and 'period'
        noise: optional noise vector with dimension (n,)

    Returns:
        Computed kernel matrix between X and Z
    """
    d = X[:, None] - Z[None]
    scaled_sin = jnp.sin(math.pi * d / params["period"]) / params["k_length"]
    k = params["k_scale"] * jnp.exp(-2 * (scaled_sin ** 2).sum(-1))
    if X.shape == Z.shape:
        k += add_jitter(noise, **kwargs) * jnp.eye(X.shape[0])
    return k


def get_kernel(kernel: Union[str, kernel_fn_type] = 'RBF'):
    kernel_book = {
        'RBF': RBFKernel,
        'Matern': MaternKernel,
        'Periodic': PeriodicKernel
    }
    if isinstance(kernel, str):
        try:
            kernel = kernel_book[kernel]
        except KeyError:
            print('Select one of the currently available kernels:',
                  *kernel_book.keys())
            raise
    return kernel



import sys
import numpy as onp
import jax.numpy as jnp
from numpy.testing import assert_equal, assert_

sys.path.insert(0, "../gpax/")

from gpax.utils import preprocess_sparse_image


def test_sparse_img_processing():
    img = onp.random.randn(16, 16)
    # Generate random indices
    idx = [onp.random.randint(0, 16) for _ in range(100)], [onp.random.randint(0, 16) for _ in range(100)]
    # Set these indices to zero
    img[idx] = 0
    # Test the utility function
    X, y, X_full = preprocess_sparse_image(img)
    assert_equal(X.ndim, 2)
    assert_equal(y.ndim,  1)
    assert_equal(X_full.ndim, 2)
    assert_(X.shape[0] < 16*16)
    assert_equal(X.shape[1], 2)
    assert_equal(y.shape[0], X.shape[0])
    assert_equal(X_full.shape[0], 16*16)
    assert_equal(X_full.shape[1], 2)



import sys
import pytest
import numpy as onp
import jax.numpy as jnp
from numpy.testing import assert_equal, assert_

sys.path.insert(0, "../gpax/")

from gpax.gp import ExactGP
from gpax.dkl import DKL
from gpax.vidkl import viDKL
from gpax.utils import get_keys
from gpax.acquisition import EI, UCB, UE, Thompson


@pytest.mark.parametrize("acq", [EI, UCB, UE, Thompson])
def test_acq_gp(acq):
    rng_keys = get_keys()
    X = onp.random.randn(8,)
    X_new = onp.random.randn(12,)
    y = 10 * X**2
    m = ExactGP(1, 'RBF')
    m.fit(rng_keys[0], X, y, num_warmup=100, num_samples=100)
    obj = acq(rng_keys[1], m, X_new)
    assert_(isinstance(obj, jnp.ndarray))
    assert_equal(obj.squeeze().shape, (len(X_new),))


@pytest.mark.parametrize("acq", [EI, UCB, UE, Thompson])
def test_acq_dkl(acq):
    rng_keys = get_keys()
    X = onp.random.randn(32, 36)
    y = onp.random.randn(32,)
    X_new = onp.random.randn(10, 36)
    m = viDKL(X.shape[-1])
    m.fit(rng_keys[0], X, y, num_steps=20, step_size=0.05)
    obj = acq(rng_keys[1], m, X_new)
    assert_(isinstance(obj, jnp.ndarray))
    assert_equal(obj.squeeze().shape, (len(X_new),))


@pytest.mark.parametrize("acq", [EI, UCB, UE])
def test_acq_penalty(acq):
    rng_keys = get_keys()
    X = onp.random.randn(8,)
    X_new = onp.random.randn(12,)
    y = 10 * X**2
    m = ExactGP(1, 'RBF')
    m.fit(rng_keys[0], X, y, num_warmup=100, num_samples=100)
    obj1 = acq(rng_keys[1], m, X_new)
    recent_points = onp.array(X_new[2:4] - 0.02)
    obj2 = acq(rng_keys[1], m, X_new, distance_penalty=0.1, recent_points=recent_points)
    assert_(onp.count_nonzero(obj1 - obj2) > 0)


@pytest.mark.parametrize("acq", [EI, UCB, UE])
def test_acq_penalty_indices(acq):
    rng_keys = get_keys()
    h = w = 5
    X = onp.random.randn(h*w, 16)
    y = onp.random.randn(len(X))
    indices = onp.array([(i, j) for i in range(h) for j in range(w)])
    X_new = onp.random.randn(h*w, 16)
    m = viDKL(input_dim=16)
    m.fit(rng_keys[0], X, y, num_steps=50)
    obj1 = acq(rng_keys[1], m, X_new, grid_indices=indices,
               distance_penalty=0.1, recent_points=indices[5:7])
    obj2 = acq(rng_keys[1], m, X_new)
    assert_(isinstance(obj1, jnp.ndarray))
    assert_equal(obj1.squeeze().shape, (len(X_new),))
    assert_(onp.count_nonzero(obj1 - obj2) > 0)



import sys
import pytest
import numpy as onp
import jax.numpy as jnp
import jax
import numpyro
from numpy.testing import assert_equal, assert_array_equal, assert_

sys.path.insert(0, "../gpax/")

from gpax.vigp import viGP
from gpax.utils import get_keys, enable_x64

enable_x64()


def get_dummy_data(jax_ndarray=True, unsqueeze=False):
    X = onp.linspace(1, 2, 8) + 0.1 * onp.random.randn(8,)
    y = (10 * X**2)
    if unsqueeze:
        X = X[:, None]
    if jax_ndarray:
        return jnp.array(X), jnp.array(y)
    return X, y


def dummy_mean_fn(x, params):
    return params["a"] * x**params["b"]


def dummy_mean_fn_priors():
    a = numpyro.sample("a", numpyro.distributions.LogNormal(0, 1))
    b = numpyro.sample("b", numpyro.distributions.Normal(3, 1))
    return {"a": a, "b": b}


def gp_kernel_custom_prior():
    length = numpyro.sample("k_length", numpyro.distributions.Uniform(0, 1))
    scale = numpyro.sample("k_scale", numpyro.distributions.LogNormal(0, 1))
    return {"k_length": length, "k_scale": scale}


@pytest.mark.parametrize("jax_ndarray", [True, False])
@pytest.mark.parametrize("unsqueeze", [True, False])
@pytest.mark.parametrize("kernel", ['RBF', 'Matern', 'Periodic'])
def test_fit(kernel, jax_ndarray, unsqueeze):
    rng_key = get_keys()[0]
    X, y = get_dummy_data(jax_ndarray, unsqueeze)
    m = viGP(1, kernel)
    m.fit(rng_key, X, y, num_steps=100)
    assert m.svi is not None


@pytest.mark.parametrize("jax_ndarray", [True, False])
@pytest.mark.parametrize("kernel", ['RBF', 'Matern', 'Periodic'])
def test_get_samples(kernel, jax_ndarray):
    rng_key = get_keys()[0]
    X, y = get_dummy_data(jax_ndarray)
    m = viGP(1, kernel)
    m.fit(rng_key, X, y, num_steps=100)
    samples = m.get_samples()
    assert isinstance(samples, dict)
    for k, v in samples.items():
        assert isinstance(k, str)
        assert isinstance(v, jnp.ndarray)


@pytest.mark.parametrize("unsqueeze", [True, False])
def test_prediction(unsqueeze):
    rng_keys = get_keys()
    X, y = get_dummy_data(unsqueeze=True)
    X_test, _ = get_dummy_data(unsqueeze=unsqueeze)
    params = {"k_length": jax.random.normal(rng_keys[0], shape=(1, 1)),
              "k_scale": jax.random.normal(rng_keys[0], shape=(1,)),
               "noise": jax.random.normal(rng_keys[0], shape=(1,))}
    m = viGP(1, 'RBF')
    m.X_train = X
    m.y_train = y
    y_mean, y_var = m.predict(rng_keys[1], X_test, params)
    assert isinstance(y_mean, jnp.ndarray)
    assert isinstance(y_var, jnp.ndarray)
    assert_equal(y_mean.shape, X_test.squeeze().shape)
    assert_equal(y_var.shape, X_test.squeeze().shape)


def test_noiseless_prediction():
    rng_keys = get_keys()
    X, y = get_dummy_data(unsqueeze=True)
    X_test, _ = get_dummy_data(unsqueeze=True)
    samples = {"k_length": jax.random.normal(rng_keys[0], shape=(1, 1)),
               "k_scale": jax.random.normal(rng_keys[0], shape=(1,)),
               "noise": jax.random.normal(rng_keys[0], shape=(1,))}
    m = viGP(1, 'RBF')
    m.X_train = X
    m.y_train = y
    y_mean1, y_var1 = m.predict(rng_keys[1], X_test, samples, noiseless=True)
    y_mean2, y_var2 = m.predict(rng_keys[1], X_test, samples, noiseless=False)
    assert_array_equal(y_mean1, y_mean2)
    assert onp.count_nonzero(y_var1 - y_var2) > 0


@pytest.mark.parametrize("unsqueeze", [True, False])
@pytest.mark.parametrize("batch_size", [2, 3, 8])
def test_prediction_in_batches(unsqueeze, batch_size):
    rng_keys = get_keys()
    X, y = get_dummy_data(unsqueeze=True)
    X_test, _ = get_dummy_data(unsqueeze=unsqueeze)
    params = {"k_length": jax.random.normal(rng_keys[0], shape=(1, 1)),
              "k_scale": jax.random.normal(rng_keys[0], shape=(1,)),
              "noise": jax.random.normal(rng_keys[0], shape=(1,))}
    m = viGP(1, 'RBF')
    m.X_train = X
    m.y_train = y
    y_mean, y_var = m.predict_in_batches(
        rng_keys[1], X_test, batch_size, params)
    assert isinstance(y_mean, jnp.ndarray)
    assert isinstance(y_var, jnp.ndarray)
    assert_equal(y_mean.shape, X_test.squeeze().shape)
    assert_equal(y_var.shape, X_test.squeeze().shape)


@pytest.mark.parametrize("kernel", ['RBF', 'Matern', 'Periodic'])
def test_fit_predict(kernel):
    rng_keys = get_keys()
    X, y = get_dummy_data()
    X_test, _ = get_dummy_data()
    m = viGP(1, kernel)
    m.fit(rng_keys[0], X, y, num_steps=100)
    y_pred, y_var = m.predict(rng_keys[1], X_test)
    assert isinstance(y_pred, jnp.ndarray)
    assert isinstance(y_var, jnp.ndarray)
    assert_equal(y_pred.shape, X_test.squeeze().shape)
    assert_equal(y_var.shape, X_test.squeeze().shape)


@pytest.mark.parametrize("batch_size", [2, 3, 8])
def test_fit_predict_in_batches(batch_size):
    rng_keys = get_keys()
    X, y = get_dummy_data()
    X_test, _ = get_dummy_data()
    m = viGP(1, "RBF")
    m.fit(rng_keys[0], X, y, num_steps=100)
    y_mean, y_var = m.predict_in_batches(
        rng_keys[1], X_test, batch_size)
    assert isinstance(y_mean, jnp.ndarray)
    assert isinstance(y_var, jnp.ndarray)
    assert_equal(y_mean.shape, X_test.squeeze().shape)
    assert_equal(y_var.shape, X_test.squeeze().shape)



@pytest.mark.parametrize("jax_ndarray", [True, False])
def test_fit_with_mean_fn(jax_ndarray):
    rng_key = get_keys()[0]
    X, y = get_dummy_data(jax_ndarray)
    m = viGP(1, 'RBF', mean_fn = lambda x: 8*x**2)
    m.fit(rng_key, X, y, num_steps=100)
    assert m.svi is not None


@pytest.mark.parametrize("jax_ndarray", [True, False])
def test_fit_with_prob_mean_fn(jax_ndarray):
    rng_key = get_keys()[0]
    X, y = get_dummy_data(jax_ndarray)
    m = viGP(1, 'RBF', mean_fn=dummy_mean_fn, mean_fn_prior=dummy_mean_fn_priors)
    m.fit(rng_key, X, y, num_steps=100)
    assert m.svi is not None


def test_fit_predict_with_mean_fn():
    rng_keys = get_keys()
    X, y = get_dummy_data()
    X_test, _ = get_dummy_data()
    m = viGP(1, 'RBF', mean_fn = lambda x: 8*x**2)
    m.fit(rng_keys[0], X, y, num_steps=100)
    y_pred, y_var = m.predict(rng_keys[1], X_test)
    assert isinstance(y_pred, jnp.ndarray)
    assert isinstance(y_var, jnp.ndarray)
    assert_equal(y_pred.shape, X_test.squeeze().shape)
    assert_equal(y_var.shape, X_test.squeeze().shape)


def test_fit_predict_with_prob_mean_fn():
    rng_keys = get_keys()
    X, y = get_dummy_data()
    X_test, _ = get_dummy_data()
    m = viGP(1, 'RBF', mean_fn=dummy_mean_fn, mean_fn_prior=dummy_mean_fn_priors)
    m.fit(rng_keys[0], X, y, num_steps=100)
    y_pred, y_var = m.predict(rng_keys[1], X_test)
    assert isinstance(y_pred, jnp.ndarray)
    assert isinstance(y_var, jnp.ndarray)
    assert_equal(y_pred.shape, X_test.squeeze().shape)
    assert_equal(y_var.shape, X_test.squeeze().shape)


def test_sample_from_prior():
    rng_key, _ = get_keys()
    X, _ = get_dummy_data()
    m = viGP(1, 'RBF')
    prior_pred = m.sample_from_prior(rng_key, X, num_samples=8)
    assert_equal(prior_pred.shape, (8, X.shape[0]))


def test_jitter_fit():
    rng_key, _ = get_keys()
    X, y = get_dummy_data()
    m1 = viGP(1, 'RBF')
    m1.fit(rng_key, X, y, num_steps=100, jitter=1e-6)
    samples1 = m1.get_samples()
    m2 = viGP(1, 'RBF')
    m2.fit(rng_key, X, y, num_steps=100, jitter=1e-6)
    samples1a = m2.get_samples()
    m3 = viGP(1, 'RBF')
    m3.fit(rng_key, X, y, num_steps=100, jitter=1e-4)
    samples2 = m3.get_samples()
    assert_(samples1["k_length"] - samples1a["k_length"] == 0)
    assert_(samples1["k_length"] - samples2["k_length"] != 0)


def test_jitter_predict():
    rng_keys = get_keys()
    X, y = get_dummy_data(unsqueeze=True)
    X_test, _ = get_dummy_data(unsqueeze=True)
    samples = {"k_length": jax.random.normal(rng_keys[0], shape=(1, 1)),
               "k_scale": jax.random.normal(rng_keys[0], shape=(1,)),
               "noise": jax.random.normal(rng_keys[0], shape=(1,))}
    m = viGP(1, 'RBF')
    m.X_train = X
    m.y_train = y
    y_mean1, y_var1 = m.predict(rng_keys[1], X_test, samples, jitter=1e-6)
    y_mean2, y_var2 = m.predict(rng_keys[1], X_test, samples, jitter=1e-4)
    assert_(onp.count_nonzero(y_var1 - y_var2) > 0)
    assert_(onp.count_nonzero(y_mean1 - y_mean2) > 0)


def test_guide_type():
    rng_key, _ = get_keys()
    X, y = get_dummy_data()
    m = viGP(1, 'RBF',  guide='delta')
    m.fit(rng_key, X, y, num_steps=100,)
    samples1 = m.get_samples()
    m = viGP(1, 'RBF',  guide='delta')
    m.fit(rng_key, X, y, num_steps=100)
    samples1a = m.get_samples()
    m = viGP(1, 'RBF',  guide='normal')
    m.fit(rng_key, X, y, num_steps=100)
    samples2 = m.get_samples()
    assert_(samples1["k_length"] - samples1a["k_length"] == 0)
    assert_(samples1["k_length"] - samples2["k_length"] != 0)



"""
vidkl.py
========

Variational inference-based implementation of deep kernel learning

Created by Maxim Ziatdinov (email: maxim.ziatdinov@ai4microscopy.com)
"""

from functools import partial
from typing import Callable, Dict, Optional, Tuple

import jax
import jax.numpy as jnp
import numpyro
import numpyro.distributions as dist
from numpyro.infer import SVI, Trace_ELBO
from numpyro.infer.autoguide import AutoDelta, AutoNormal
from numpyro.contrib.module import random_haiku_module
from jax import jit
import haiku as hk

from .gp import ExactGP
from .kernels import get_kernel
from .utils import get_haiku_dict


class viDKL(ExactGP):
    """
    Implementation of the variational infernece-based deep kernel learning

    Args:
        input_dim:
            Number of input dimensions
        z_dim:
            Latent space dimensionality (defaults to 2)
        kernel:
            Kernel function ('RBF', 'Matern', 'Periodic', or custom function)
        kernel_prior:
            Optional priors over kernel hyperparameters; uses LogNormal(0,1) by default
        nn:
            Custom neural network ('feature extractor'); uses a 3-layer MLP
            with ReLU activations by default
        latent_prior:
            Optional prior over the latent space (NN embedding); uses none by default
        guide:
            Auto-guide option, use 'delta' (default) or 'normal'
    
    Examples:

        vi-DKL with image patches as inputs and a 1-d vector as targets

        >>> # Get random number generator keys for training and prediction
        >>> key1, key2 = gpax.utils.get_keys()
        >>> input data dimensions are (n, height*width*channels)
        >>> data_dim = X.shape[-1]
        >>> # Initialize vi-DKL model with 2 latent dimensions
        >>> dkl = gpax.viDKL(input_dim=data_dim, z_dim=2, kernel='RBF')
        >>> Train a model
        >>> dkl.fit(rng_key, X_train, y_train, num_steps=1000, step_size=0.005)
        >>> # Obtain posterior mean and variance ('uncertainty') at new inputs
        >>> y_mean, y_var = dkl.predict(key2, X_new)
    """

    def __init__(self, input_dim: int, z_dim: int = 2, kernel: str = 'RBF',
                 kernel_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 nn: Optional[Callable[[jnp.ndarray], jnp.ndarray]] = None,
                 latent_prior: Optional[Callable[[jnp.ndarray], Dict[str, jnp.ndarray]]] = None,
                 guide: str = 'delta'
                 ) -> None:
        super(viDKL, self).__init__(input_dim, kernel, None, kernel_prior)
        if guide not in ['delta', 'normal']:
            raise NotImplementedError("Select guide between 'delta' and 'normal'")
        nn_module = nn if nn else MLP
        self.nn_module = hk.transform(lambda x: nn_module(z_dim)(x))
        self.kernel_dim = z_dim
        self.data_dim = (input_dim,) if isinstance(input_dim, int) else input_dim
        self.latent_prior = latent_prior
        self.guide_type = AutoNormal if guide == 'normal' else AutoDelta
        self.kernel_params = None
        self.nn_params = None

    def model(self, X: jnp.ndarray, y: jnp.ndarray = None, **kwargs) -> None:
        """DKL probabilistic model"""
        # NN part
        feature_extractor = random_haiku_module(
            "feature_extractor", self.nn_module, input_shape=(1, *self.data_dim),
            prior=(lambda name, shape: dist.Cauchy() if name.startswith("b") else dist.Normal()))
        z = feature_extractor(X)
        if self.latent_prior:  # Sample latent variable
            z = self.latent_prior(z)
        # Sample GP kernel parameters
        if self.kernel_prior:
            kernel_params = self.kernel_prior()
        else:
            kernel_params = self._sample_kernel_params()
        # Sample noise
        noise = numpyro.sample("noise", dist.LogNormal(0.0, 1.0))
        # GP's mean function
        f_loc = jnp.zeros(z.shape[0])
        # compute kernel
        k = get_kernel(self.kernel)(
            z, z,
            kernel_params,
            noise,
            **kwargs
        )
        # sample y according to the standard Gaussian process formula
        numpyro.sample(
            "y",
            dist.MultivariateNormal(loc=f_loc, covariance_matrix=k),
            obs=y,
        )

    def single_fit(self, rng_key: jnp.array, X: jnp.ndarray, y: jnp.ndarray,
                   num_steps: int = 1000, step_size: float = 5e-3,
                   print_summary: bool = True, progress_bar=True,
                   **kwargs) -> None:
        """
        Optimizes parameters of a single DKL model
        """
        # Setup optimizer and SVI
        optim = numpyro.optim.Adam(step_size=step_size, b1=0.5)
        svi = SVI(
            self.model,
            guide=self.guide_type(self.model),
            optim=optim,
            loss=Trace_ELBO(),
            X=X,
            y=y,
            **kwargs
        )
        params, _, losses = svi.run(rng_key, num_steps, progress_bar=progress_bar)
        # Get DKL parameters from the guide
        params_map = svi.guide.median(params)
        # Get NN weights
        nn_params = get_haiku_dict(params_map)
        # Get GP kernel hyperparmeters
        kernel_params = {k: v for (k, v) in params_map.items()
                         if not k.startswith("feature_extractor")}
        return nn_params, kernel_params, losses

    def fit(self, rng_key: jnp.array, X: jnp.ndarray, y: jnp.ndarray,
            num_steps: int = 1000, step_size: float = 5e-3,
            print_summary: bool = True, progress_bar=True, **kwargs):
        """
        Run stochastic variational inference to learn a DKL model(s) parameters

        Args:
            rng_key: random number generator key
            X: Input high-dimensional features
            y: Target output (scalar of vector)
            num_steps: number of SVI steps
            step_size: step size schedule for Adam optimizer
            print_summary: print summary at the end of sampling
            progress_bar: show progress bar (works only for scalar outputs)
        """
        def _single_fit(x_i, y_i):
            return self.single_fit(
                rng_key, x_i, y_i, num_steps, step_size,
                print_summary=False, progress_bar=False, **kwargs)

        self.X_train = X
        self.y_train = y

        if X.ndim == len(self.data_dim) + 2:
            self.nn_params, self.kernel_params, self.loss = jax.vmap(_single_fit)(X, y)
            if progress_bar:
                avg_bw = [num_steps - num_steps // 20, num_steps]
                print("init loss: {}, final loss (avg) [{}-{}]: {} ".format(
                    self.loss[0].mean(), avg_bw[0], avg_bw[1],
                    self.loss.mean(0)[avg_bw[0]:avg_bw[1]].mean().round(4)))
        else:
            self.nn_params, self.kernel_params, self.loss = self.single_fit(
                rng_key, X, y, num_steps, step_size, print_summary, progress_bar
            )
        if print_summary:
            self._print_summary()

    @partial(jit, static_argnames='self')
    def get_mvn_posterior(self,
                          X_train: jnp.ndarray,
                          y_train: jnp.ndarray,
                          X_new: jnp.ndarray,
                          nn_params: Dict[str, jnp.ndarray],
                          k_params: Dict[str, jnp.ndarray],
                          noiseless: bool = False,
                          **kwargs
                          ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """
        Returns predictive mean and covariance at new points
        (mean and cov, where cov.diagonal() is 'uncertainty')
        given a single set of DKL parameters
        """
        noise = k_params["noise"]
        noise_p = noise * (1 - jnp.array(noiseless, int))
        # embed data into the latent space
        z_train = self.nn_module.apply(
            nn_params, jax.random.PRNGKey(0), X_train)
        z_test = self.nn_module.apply(
            nn_params, jax.random.PRNGKey(0), X_new)
        # compute kernel matrices for train and test data
        k_pp = get_kernel(self.kernel)(z_test, z_test, k_params, noise_p, **kwargs)
        k_pX = get_kernel(self.kernel)(z_test, z_train, k_params, jitter=0.0)
        k_XX = get_kernel(self.kernel)(z_train, z_train, k_params, noise, **kwargs)
        # compute the predictive covariance and mean
        K_xx_inv = jnp.linalg.inv(k_XX)
        cov = k_pp - jnp.matmul(k_pX, jnp.matmul(K_xx_inv, jnp.transpose(k_pX)))
        mean = jnp.matmul(k_pX, jnp.matmul(K_xx_inv, y_train))
        return mean, cov

    def sample_from_posterior(self, rng_key: jnp.ndarray,
                              X_new: jnp.ndarray, n: int = 1000,
                              noiseless: bool = False,
                              **kwargs
                              ) -> Tuple[jnp.ndarray]:
        """
        Samples from the DKL posterior at X_new points
        """
        y_mean, K = self.get_mvn_posterior(
            self.X_train, self.y_train, X_new,
            self.nn_params, self.kernel_params, noiseless, **kwargs)
        y_sampled = dist.MultivariateNormal(y_mean, K).sample(rng_key, sample_shape=(n,))
        return y_mean, y_sampled

    def predict_in_batches(self, rng_key: jnp.ndarray,
                           X_new: jnp.ndarray,  batch_size: int = 100,
                           noiseless: bool = False,
                           **kwargs
                           ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """
        Make prediction at X_new with sampled DKL parameters
        by spitting the input array into chunks ("batches") and running
        self.predict on each of them one-by-one to avoid a memory overflow
        """
        predict_fn = lambda xi: self.predict(rng_key, xi, noiseless=noiseless, **kwargs)
        cat_dim = 1 if self.X_train.ndim == len(self.data_dim) + 2 else 0
        mean, var = self._predict_in_batches(
            rng_key, X_new, batch_size, cat_dim, predict_fn=predict_fn)
        mean = jnp.concatenate(mean, cat_dim)
        var = jnp.concatenate(var, cat_dim)
        return mean, var

    def predict(self, rng_key: jnp.ndarray, X_new: jnp.ndarray,
                params: Optional[Tuple[Dict[str, jnp.ndarray]]] = None,
                noiseless: bool = False, *args, **kwargs
                ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """
        Make prediction at X_new points using a trained DKL model(s)

        Args:
            rng_key: random number generator key
            X_new: New inputs
            params: Tuple with neural network weigths and kernel parameters (optional)
            noiseless:
                Noise-free prediction. It is set to False by default as new/unseen data is assumed
                to follow the same distribution as the training data. Hence, since we introduce a model noise
                for the training data, we also want to include that noise in our prediction.

        Returns:
            Predictive mean and variance
        """

        def single_predict(x_train_i, y_train_i, x_new_i, nnpar_i, kpar_i):
            mean, cov = self.get_mvn_posterior(
                x_train_i, y_train_i, x_new_i, nnpar_i, kpar_i, noiseless, **kwargs)
            return mean, cov.diagonal()

        if params is None:
            nn_params = self.nn_params
            k_params = self.kernel_params
        else:
            nn_params, k_params = params

        p_args = (self.X_train, self.y_train, X_new, nn_params, k_params)
        if self.X_train.ndim == len(self.data_dim) + 2:
            mean, var = jax.vmap(single_predict)(*p_args)
        else:
            mean, var = single_predict(*p_args)

        return mean, var

    def fit_predict(self, rng_key: jnp.array, X: jnp.ndarray, y: jnp.ndarray,
                    X_new: jnp.ndarray, num_steps: int = 1000, step_size: float = 5e-3,
                    n_models: int = 1, batch_size: int = 100, noiseless: bool = False,
                    ensemble_method: str = 'vectorized',
                    print_summary: bool = True, progress_bar=True, **kwargs
                    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """
        Run SVI to learn DKL model(s) parameters and make a prediction with
        trained model(s) on new data. Allows using an ensemble of models.

        Args:
            rng_key: random number generator key
            X: Input high-dimensional features
            y: Target output (scalar of vector)
            X_new: New ('test') data
            num_steps: number of SVI steps
            step_size: step size schedule for Adam optimizer
            n_models: number of models in the ensemble (defaults to 1)
            batch_size: prediction batch size (to avoid memory overflows)
            noiseless:
                Noise-free prediction. It is set to False by default as new/unseen data is assumed
                to follow the same distribution as the training data. Hence, since we introduce a model noise
                for the training data, we also want to include that noise in our prediction.
            ensemble_method: 'vectorized' (single GPU) or 'parallel' (multiple GPUs)
            print_summary: print summary at the end of sampling
            progress_bar: show progress bar (works only for scalar outputs)

        Returns:
            Predictive mean and variance
        """

        def single_fit_predict(key):
            self.fit(key, X, y, num_steps, step_size,
                     print_summary, progress_bar, **kwargs)
            mean, var = self.predict_in_batches(
                key, X_new, batch_size, noiseless, **kwargs)
            return mean, var

        if n_models > 1 and ensemble_method not in ["vectorized", "parallel"]:
            raise ValueError(
                "For the ensemble_method, select between 'vectorized and 'parallel'.")
        keys = jax.random.split(rng_key, num=n_models)
        if n_models > 1:
            pstrategy = jax.vmap if ensemble_method == 'vectorized' else jax.pmap
            print_summary = progress_bar = 0
            mean, var = pstrategy(single_fit_predict)(keys)
        else:
            mean, var = single_fit_predict(keys[0])

        return mean, var

    @partial(jit, static_argnames='self')
    def embed(self, X_new: jnp.ndarray) -> jnp.ndarray:
        """
        Use trained neural network(s) to embed the input data
        into the latent space(s)
        """
        def single_embed(nnpar_i, x_i):
            return self.nn_module.apply(nnpar_i, jax.random.PRNGKey(0), x_i)

        if self.X_train.ndim == len(self.data_dim) + 2:
            z = jax.vmap(single_embed)(self.nn_params, X_new)
        else:
            z = single_embed(self.nn_params, X_new)
        return z

    def _print_summary(self) -> None:
        if isinstance(self.kernel_params, dict):
            print('\nInferred GP kernel parameters')
            if self.X_train.ndim == len(self.data_dim) + 1:
                for (k, vals) in self.kernel_params.items():
                    spaces = " " * (15 - len(k))
                    print(k, spaces, jnp.around(vals, 4))
            else:
                for (k, vals) in self.kernel_params.items():
                    for i, v in enumerate(vals):
                        spaces = " " * (15 - len(k))
                        print(k+"[{}]".format(i), spaces, jnp.around(v, 4))


class MLP(hk.Module):
    """Simple MLP"""
    def __init__(self, embedim=2):
        super().__init__()
        self._embedim = embedim

    def __call__(self, x):
        x = hk.Linear(64)(x)
        x = jax.nn.relu(x)
        x = hk.Linear(64)(x)
        x = jax.nn.relu(x)
        x = hk.Linear(self._embedim)(x)
        return x



import sys
import pytest
import numpy as onp
import jax.numpy as jnp
import jax
import numpyro
from numpy.testing import assert_equal, assert_array_equal

sys.path.insert(0, "../gpax/")

from gpax.spm import sPM
from gpax.utils import get_keys


def get_dummy_data(jax_ndarray=True):
    X = onp.linspace(1, 2, 8) + 0.1 * onp.random.randn(8,)
    y = (10 * X**2)
    if jax_ndarray:
        return jnp.array(X), jnp.array(y)
    return X, y


def model(x, params):
    return params["a"] * x**params["b"]


def model_priors():
    a = numpyro.sample("a", numpyro.distributions.LogNormal(0, 1))
    b = numpyro.sample("b", numpyro.distributions.Normal(3, 1))
    return {"a": a, "b": b}


@pytest.mark.parametrize("jax_ndarray", [True, False])
def test_fit(jax_ndarray):
    rng_key = get_keys()[0]
    X, y = get_dummy_data(jax_ndarray)
    m = sPM(model, model_priors)
    m.fit(rng_key, X, y, num_warmup=100, num_samples=100)
    assert m.mcmc is not None


def test_get_samples():
    rng_key = get_keys()[0]
    X, y = get_dummy_data()
    m = sPM(model, model_priors)
    m.fit(rng_key, X, y, num_warmup=100, num_samples=100)
    samples = m.get_samples()
    assert isinstance(samples, dict)
    for k, v in samples.items():
        assert isinstance(k, str)
        assert isinstance(v, jnp.ndarray)
        assert_equal(len(v), 100)


def test_prediction():
    rng_keys = get_keys()
    X, y = get_dummy_data()
    X_test, _ = get_dummy_data()
    samples = {"a": jax.random.normal(rng_keys[0], shape=(100, 1)),
               "b": jax.random.normal(rng_keys[0], shape=(100,))}
    m =sPM(model, model_priors)
    y_mean, y_sampled = m.predict(rng_keys[1], X_test, samples)
    assert isinstance(y_mean, jnp.ndarray)
    assert isinstance(y_sampled, jnp.ndarray)
    assert_equal(y_mean.shape, X_test.squeeze().shape)
    assert_equal(y_sampled.shape, (100, X_test.shape[0]))


import sys
import pytest
import numpy as onp
import jax.numpy as jnp
import jax
import numpyro
from numpy.testing import assert_equal, assert_array_equal, assert_

sys.path.insert(0, "../gpax/")

from gpax.vgp import vExactGP
from gpax.utils import get_keys


def get_dummy_data(jax_ndarray=True, unsqueeze=False):
    get_data = lambda: onp.linspace(1, 2, 8) + 0.1 * onp.random.randn(8,)
    X = onp.array([get_data() for _ in range(3)])
    y = (10 * X**2)
    if unsqueeze:
        X = X[..., None]
    if jax_ndarray:
        return jnp.array(X), jnp.array(y)
    return X, y


@pytest.mark.parametrize("jax_ndarray", [True, False])
@pytest.mark.parametrize("unsqueeze", [True, False])
@pytest.mark.parametrize("kernel", ['RBF', 'Matern', 'Periodic'])
def test_fit(kernel, jax_ndarray, unsqueeze):
    rng_key = get_keys()[0]
    X, y = get_dummy_data(jax_ndarray, unsqueeze)
    m = vExactGP(1, kernel)
    m.fit(rng_key, X, y, num_warmup=50, num_samples=50)
    assert m.mcmc is not None


@pytest.mark.parametrize("jax_ndarray", [True, False])
@pytest.mark.parametrize("kernel", ['RBF', 'Matern', 'Periodic'])
def test_get_samples(kernel, jax_ndarray):
    rng_key = get_keys()[0]
    X, y = get_dummy_data(jax_ndarray)
    m = vExactGP(1, kernel)
    m.fit(rng_key, X, y, num_warmup=50, num_samples=50)
    samples = m.get_samples()
    assert isinstance(samples, dict)
    for k, v in samples.items():
        assert isinstance(k, str)
        assert isinstance(v, jnp.ndarray)
        assert_equal(v.shape[:2], (50, 3))


@pytest.mark.parametrize("chain_dim, samples_dim", [(True, 3), (False, 2)])
def test_get_samples_chain_dim(chain_dim, samples_dim):
    rng_key = get_keys()[0]
    X, y = get_dummy_data()
    m = vExactGP(1, 'RBF')
    m.fit(rng_key, X, y, num_warmup=50, num_samples=50, num_chains=2)
    samples = m.get_samples(chain_dim)
    assert_equal(samples["k_scale"].ndim, samples_dim)
    assert_equal(samples["noise"].ndim, samples_dim)
    assert_equal(samples["k_length"].ndim, samples_dim + 1)


@pytest.mark.parametrize("kernel", ['RBF', 'Matern'])
def test_sample_kernel(kernel):
    m = vExactGP(1, kernel)
    with numpyro.handlers.seed(rng_seed=1):
        kernel_params = m._sample_kernel_params(task_dim=3)
    _ = kernel_params.pop('period')
    param_names = ['k_length', 'k_scale']
    for k, v in kernel_params.items():
        assert k in param_names
        assert isinstance(v, jnp.ndarray)


def test_sample_periodic_kernel():
    m = vExactGP(1, 'Periodic')
    with numpyro.handlers.seed(rng_seed=1):
        kernel_params = m._sample_kernel_params(task_dim=3)
    param_names = ['k_length', 'k_scale', 'period']
    for k, v in kernel_params.items():
        assert k in param_names
        assert isinstance(v, jnp.ndarray)


def test_get_mvn_posterior():
    X, y = get_dummy_data(unsqueeze=True)
    X_test, _ = get_dummy_data(unsqueeze=True)
    params = {"k_length": jnp.array([[1.0], [1.0], [1.0]]),
              "k_scale": jnp.array([1.0, 1.0, 1.0]),
              "noise": jnp.array([0.1, 0.1, 0.1])}
    m = vExactGP(1, 'RBF')
    m.X_train = X
    m.y_train = y
    mean, cov = m.get_mvn_posterior(X_test, params)
    assert isinstance(mean, jnp.ndarray)
    assert isinstance(cov, jnp.ndarray)
    assert_equal(mean.shape, X_test.shape[:-1])
    assert_equal(cov.shape, (X_test.shape[0], X_test.shape[1], X_test.shape[1]))


def test_get_mvn_posterior_noiseless():
    X, y = get_dummy_data(unsqueeze=True)
    X_test, _ = get_dummy_data(unsqueeze=True)
    params = {"k_length": jnp.array([[1.0], [1.0], [1.0]]),
              "k_scale": jnp.array([1.0, 1.0, 1.0]),
              "noise": jnp.array([0.1, 0.1, 0.1])}
    m = vExactGP(1, 'RBF')
    m.X_train = X
    m.y_train = y
    mean1, cov1 = m.get_mvn_posterior(X_test, params, noiseless=False)
    mean1_, cov1_ = m.get_mvn_posterior(X_test, params, noiseless=False)
    mean2, cov2 = m.get_mvn_posterior(X_test, params, noiseless=True)
    assert_array_equal(mean1, mean1_)
    assert_array_equal(cov1, cov1_)
    assert_array_equal(mean1, mean2)
    assert onp.count_nonzero(cov1 - cov2) > 0


@pytest.mark.parametrize("n", [1, 10])
def test_prediction(n):
    rng_keys = get_keys()
    X, y = get_dummy_data(unsqueeze=True)
    X_test, _ = get_dummy_data(unsqueeze=True)
    samples = {"k_length": jax.random.normal(rng_keys[0], shape=(100, 3, 1)),
               "k_scale": jax.random.normal(rng_keys[0], shape=(100, 3)),
               "noise": jax.random.normal(rng_keys[0], shape=(100, 3))}
    m = vExactGP(1, 'RBF')
    m.X_train = X
    m.y_train = y
    y_mean, y_sampled = m.predict(rng_keys[1], X_test, samples, n=n)
    assert isinstance(y_mean, jnp.ndarray)
    assert isinstance(y_sampled, jnp.ndarray)
    assert_equal(y_mean.shape, X_test.shape[:-1])
    assert_equal(y_sampled.shape, (100, n, *X_test.shape[:-1]))


@pytest.mark.parametrize("n", [1, 10])
def test_fit_predict_in_batches(n):
    rng_keys = get_keys()
    X, y = get_dummy_data()
    X_test, _ = get_dummy_data()
    m = vExactGP(1, 'RBF')
    m.fit(rng_keys[0], X, y, num_warmup=50, num_samples=50)
    y_pred, y_sampled = m.predict_in_batches(rng_keys[1], X_test, batch_size=4, n=n)
    assert isinstance(y_pred, jnp.ndarray)
    assert isinstance(y_sampled, jnp.ndarray)
    assert_equal(y_pred.shape, X_test.squeeze().shape)
    print(y_sampled.shape)
    assert_equal(y_sampled.shape, (50, n, *X_test.shape))


@pytest.mark.parametrize("n", [1, 10])
def test_fit_predict_in_batches_noiseless(n):
    rng_keys = get_keys()
    X, y = get_dummy_data()
    X_test, _ = get_dummy_data()
    m = vExactGP(1, 'RBF')
    m.fit(rng_keys[0], X, y, num_warmup=50, num_samples=50)
    y_mean1, y_sampled1 = m.predict_in_batches(rng_keys[1], X_test, batch_size=4, n=n, noiseless=True)
    y_mean2, y_sampled2 = m.predict_in_batches(rng_keys[1], X_test, batch_size=4, n=n, noiseless=False)
    assert_array_equal(y_mean1, y_mean2)
    assert onp.count_nonzero(y_sampled1 - y_sampled2) > 0


def test_jitter_predict():
    rng_keys = get_keys()
    X, y = get_dummy_data(unsqueeze=True)
    X_test, _ = get_dummy_data(unsqueeze=True)
    samples = {"k_length": jax.random.normal(rng_keys[0], shape=(100, 3, 1)),
               "k_scale": jax.random.normal(rng_keys[0], shape=(100, 3)),
               "noise": jax.random.normal(rng_keys[0], shape=(100, 3))}
    m = vExactGP(1, 'RBF')
    m.X_train = X
    m.y_train = y
    y_mean1, y_sampled1 = m.predict(rng_keys[1], X_test, samples, n=1, jitter=1e-6)
    y_mean2, y_sampled2 = m.predict(rng_keys[1], X_test, samples, n=1, jitter=1e-5)
    assert_(onp.count_nonzero(y_sampled1 - y_sampled2) > 0)
    assert_(onp.count_nonzero(y_mean1 - y_mean2) > 0)



"""
acquisition.py
==============

Acquisition functions

Created by Maxim Ziatdinov (email: maxim.ziatdinov@ai4microscopy.com)
"""

from typing import Type, Tuple, Optional

import jax
import jax.numpy as jnp
import jax.random as jra
import numpy as onp
import numpyro.distributions as dist

from .gp import ExactGP
from .vidkl import viDKL


def EI(rng_key: jnp.ndarray, model: Type[ExactGP],
       X: jnp.ndarray, xi: float = 0.01,
       maximize: bool = False, n: int = 1,
       noiseless: bool = False, distance_penalty: float = None,
       recent_points: jnp.ndarray = None, grid_indices: jnp.ndarray = None,
       **kwargs) -> jnp.ndarray:
    r"""
    Expected Improvement

    Args:
        rng_key: JAX random number generator key
        model: trained model
        X: new inputs
        maximize: If True, assumes that BO is solving maximization problem
        n: number of samples drawn from each MVN distribution
           (number of distributions is equal to the number of HMC samples)
        noiseless:
            Noise-free prediction. It is set to False by default as new/unseen data is assumed
            to follow the same distribution as the training data. Hence, since we introduce a model noise
            for the training data, we also want to include that noise in our prediction.
        distance_penalty:
            Modifies the acquisition function by penalizing points near the recent points as

            .. math::
                \alpha - \lambda \cdot \pi(X, r)

            where :math:`\pi(X, r)` computes a penalty for points in :math:`X` based on their distance to recent points `r`, 
            :math:`\alpha` represents the acquisition function, and :math:`\lambda` represents the distance penalty. Defaults to None.
        recent_points:
            An array of recently visited points [oldest, ..., newest] provided by user
        grid_indices:
            Grid indices of data points in X array for the penalty term calculation.
            For example, if each data point is an image patch, the indices could correspond
            to the (i, j) pixel coordinates of their centers in the original image.
        **jitter:
            Small positive term added to the diagonal part of a covariance
            matrix for numerical stability (Default: 1e-6)
    """
    X = X[:, None] if X.ndim < 2 else X
    if model.mcmc is not None:
        y_mean, y_sampled = model.predict(
            rng_key, X, n=n, noiseless=noiseless, **kwargs)
        y_sampled = y_sampled.reshape(n * y_sampled.shape[0], -1)
        mean, sigma = y_sampled.mean(0), y_sampled.std(0)
        best_f = y_mean.max() if maximize else y_mean.min()
    else:
        mean, var = model.predict(
            rng_key, X, noiseless=noiseless, **kwargs)
        sigma = jnp.sqrt(var)
        best_f = mean.max() if maximize else mean.min()
    u = (mean - best_f) / sigma
    if not maximize:
        u = -u
    normal = dist.Normal(jnp.zeros_like(u), jnp.ones_like(u))
    ucdf = normal.cdf(u)
    updf = jnp.exp(normal.log_prob(u))
    acq = sigma * (updf + u * ucdf)
    if distance_penalty is not None:
        X_ = grid_indices if grid_indices is not None else X
        penalties = jax.vmap(penalty_point, in_axes=(0, None))(X_, recent_points)
        acq -= distance_penalty * penalties
    return acq


def UCB(rng_key: jnp.ndarray, model: Type[ExactGP],
        X: jnp.ndarray, beta: float = .25,
        maximize: bool = False, n: int = 1,
        noiseless: bool = False, distance_penalty: float = None,
        recent_points: jnp.ndarray = None, grid_indices: jnp.ndarray = None,
        **kwargs) -> jnp.ndarray:
    r"""
    Upper confidence bound

    Args:
        rng_key: JAX random number generator key
        model: trained model
        X: new inputs
        beta: coefficient balancing exploration-exploitation trade-off
        maximize: If True, assumes that BO is solving maximization problem
        n: number of samples drawn from each MVN distribution
           (number of distributions is equal to the number of HMC samples)
        noiseless:
            Noise-free prediction. It is set to False by default as new/unseen data is assumed
            to follow the same distribution as the training data. Hence, since we introduce a model noise
            for the training data, we also want to include that noise in our prediction.
        distance_penalty:
            Modifies the acquisition function by penalizing points near the recent points as

            .. math::
                \alpha - \lambda \cdot \pi(X, r)

            where :math:`\pi(X, r)` computes a penalty for points in :math:`X` based on their distance to recent points `r`, 
            :math:`\alpha` represents the acquisition function, and :math:`\lambda` represents the distance penalty. Defaults to None.
        recent_points:
            An array of recently visited points [oldest, ..., newest] provided by user
        grid_indices:
            Grid indices of data points in X array for the penalty term calculation.
            For example, if each data point is an image patch, the indices could correspond
            to the (i, j) pixel coordinates of their centers in the original image.
        **jitter:
            Small positive term added to the diagonal part of a covariance
            matrix for numerical stability (Default: 1e-6)
    """
    X = X[:, None] if X.ndim < 2 else X
    if model.mcmc is not None:
        _, y_sampled = model.predict(
            rng_key, X, n=n, noiseless=noiseless, **kwargs)
        y_sampled = y_sampled.reshape(n * y_sampled.shape[0], -1)
        mean, var = y_sampled.mean(0), y_sampled.var(0)
    else:
        mean, var = model.predict(
            rng_key, X, noiseless=noiseless, **kwargs)
    delta = jnp.sqrt(beta * var)
    if maximize:
        acq = mean + delta
    else:
        acq = delta - mean  # we return a negative acq for argmax in BO
    if distance_penalty is not None:
        X_ = grid_indices if grid_indices is not None else X
        penalties = jax.vmap(penalty_point, in_axes=(0, None))(X_, recent_points)
        acq -= distance_penalty * penalties
    return acq


def UE(rng_key: jnp.ndarray,
       model: Type[ExactGP],
       X: jnp.ndarray, n: int = 1,
       noiseless: bool = False,
       distance_penalty: float = None,
       recent_points: jnp.ndarray = None,
       grid_indices: jnp.ndarray = None,
       **kwargs) -> jnp.ndarray:
    r"""
    Uncertainty-based exploration

    Args:
        rng_key: JAX random number generator key
        model: trained model
        X: new inputs
        n: number of samples drawn from each MVN distribution
           (number of distributions is equal to the number of HMC samples)
        noiseless:
            Noise-free prediction. It is set to False by default as new/unseen data is assumed
            to follow the same distribution as the training data. Hence, since we introduce a model noise
            for the training data, we also want to include that noise in our prediction.
        distance_penalty:
            Modifies the uncertainty by penalizing points near the recent points as

            .. math::
                \sigma - \lambda \cdot \pi(X, r)

            where :math:`\pi(X, r)` computes a penalty for points in :math:`X` based on their distance to recent points `r`, 
            :math:`\sigma` represents the uncertainty, and :math:`\lambda` represents the distance penalty. Defaults to None.
        recent_points:
            An array of recently visited points [oldest, ..., newest] provided by user
        grid_indices:
            Grid indices of data points in X array for the penalty term calculation.
            For example, if each data point is an image patch, the indices could correspond
            to the (i, j) pixel coordinates of their centers in the original image.
        **jitter:
            Small positive term added to the diagonal part of a covariance
            matrix for numerical stability (Default: 1e-6)
    """
    X = X[:, None] if X.ndim < 2 else X
    if model.mcmc is not None:
        _, y_sampled = model.predict(
            rng_key, X, n=n, noiseless=noiseless, **kwargs)
        y_sampled = y_sampled.mean(1)
        var = y_sampled.var(0)
    else:
        _, var = model.predict(
            rng_key, X, noiseless=noiseless, **kwargs)
    if distance_penalty is not None:
        X_ = grid_indices if grid_indices is not None else X
        penalties = jax.vmap(penalty_point, in_axes=(0, None))(X_, recent_points)
        var -= distance_penalty * penalties
    return var


def Thompson(rng_key: jnp.ndarray,
             model: Type[ExactGP],
             X: jnp.ndarray, n: int = 1,
             noiseless: bool = False,
             **kwargs) -> jnp.ndarray:
    """
    Thompson sampling

    Args:
        rng_key: JAX random number generator key
        model: trained model
        X: new inputs
        n: number of samples drawn from the randomly selected MVN distribution
        noiseless:
            Noise-free prediction. It is set to False by default as new/unseen data is assumed
            to follow the same distribution as the training data. Hence, since we introduce a model noise
            for the training data, we also want to include that noise in our prediction.
        **jitter:
            Small positive term added to the diagonal part of a covariance
            matrix for numerical stability (Default: 1e-6)
    """
    if model.mcmc is not None:
        posterior_samples = model.get_samples()
        idx = jra.randint(rng_key, (1,), 0, len(posterior_samples["k_length"]))
        samples = {k: v[idx] for (k, v) in posterior_samples.items()}
        _, tsample = model.predict(
            rng_key, X, samples, n, noiseless=noiseless, **kwargs)
        if n > 1:
            tsample = tsample.mean(1).squeeze()
    else:
        _, tsample = model.sample_from_posterior(
            rng_key, X, n=1, noiseless=noiseless, **kwargs)
    return tsample


def qUCB(rng_key: jnp.ndarray, model: Type[ExactGP],
         X: jnp.ndarray, indices: Optional[jnp.ndarray] = None,
         qbatch_size: int = 4, alpha: float = 1.0, beta: float = .25,
         maximize: bool = True, n: int = 500,
         n_restarts: int = 20, noiseless: bool = False,
         **kwargs) -> jnp.ndarray:
    """
    The acquisition function defined as alpha * mu + sqrt(beta) * sigma
    that can output a "batch" of next points to evaluate. It takes advantage of
    the fact that in MCMC-based GP or DKL we obtain a separate multivariate
    normal posterior for each set of sampled kernel hyperparameters.

    Args:
        rng_key: random number generator key
        model: ExactGP or DKL type of model
        X: input array
        indices: indices of data points in X array. For example, if
            each data point is an image patch, the indices should
            correspond to their (x, y) coordinates in the original image.
        qbatch_size: desired number of sampled points (default: 4)
        alpha: coefficient before mean prediction term (default: 1.0)
        beta: coefficient before variance term (default: 0.25)
        maximize: sign of variance term (+/- if True/False)
        n: number of draws from each multivariate normal posterior
        n_restarts: number of restarts to find a batch of maximally
            separated points to evaluate next
        noiseless: noise-free prediction for new/test data (default: False)

    Returns:
        Computed acquisition function with qbatch x features
        or task x qbatch x features dimensions
    """
    if model.mcmc is None:
        raise NotImplementedError(
            "Currently supports only ExactGP and DKL with MCMC inference")
    dist_all, obj_all = [], []
    X_ = jnp.array(indices) if indices is not None else jnp.array(X)
    for _ in range(n_restarts):
        y_sampled = obtain_samples(
            rng_key, model, X, qbatch_size, n, noiseless, **kwargs)
        mean, var = y_sampled.mean(1), y_sampled.var(1)
        delta = jnp.sqrt(beta * var)
        if maximize:
            obj = alpha * mean + delta
            points = X_[obj.argmax(-1)]
        else:
            obj = alpha * mean - delta
            points = X_[obj.argmin(-1)]
        d = jnp.linalg.norm(points, axis=-1).mean(0)
        dist_all.append(d)
        obj_all.append(obj)
    idx = jnp.array(dist_all).argmax(0)
    if idx.ndim > 0:
        obj_all = jnp.array(obj_all)
        return jnp.array([obj_all[j,:,i] for i, j in enumerate(idx)])
    return obj_all[idx]


def obtain_samples(rng_key: jnp.ndarray, model: Type[ExactGP],
                   X: jnp.ndarray, qbatch_size: int = 4,
                   n: int = 500, noiseless: bool = False,
                   **kwargs) -> jnp.ndarray:
    xbatch_size = kwargs.get("xbatch_size", 100)
    posterior_samples = model.get_samples()
    idx = onp.arange(0, len(posterior_samples["k_length"]))
    onp.random.shuffle(idx)
    idx = idx[:qbatch_size]
    samples = {k: v[idx] for (k, v) in posterior_samples.items()}
    if X.shape[0] > xbatch_size:
        _, y_sampled = model.predict(
            rng_key, X, samples, n,
            noiseless=noiseless, **kwargs)
    else:
        _, y_sampled = model.predict_in_batches(
            rng_key, X, xbatch_size, samples, n,
            noiseless=noiseless, **kwargs)
    return y_sampled


def penalty_point(x: jnp.ndarray, recent_points: jnp.ndarray) -> jnp.ndarray:
    """
    Compute a penalty for point x based on its distance to recent points.
    """
    if x.ndim == 1:
        x = x[:, None]
    if recent_points.ndim == 1:
        recent_points = recent_points[:, None]
    distances = jnp.linalg.norm(recent_points - x, axis=1)
    # Penalties are inversely proportional to distance and timestamp
    if len(recent_points) == 1:
        timestamps = 1
    else:
        timestamps = jnp.arange(len(recent_points), 0, -1)
    penalties = 1 / distances / timestamps
    return jnp.sum(penalties)



import sys
import pytest
import numpy as onp
import jax.numpy as jnp
import jax
import haiku as hk
import numpyro
from numpy.testing import assert_equal, assert_array_equal

sys.path.insert(0, "../gpax/")

from gpax.vidkl import viDKL, MLP
from gpax.utils import get_keys


def get_dummy_data(jax_ndarray=True):
    X = onp.random.randn(21, 36)
    y = onp.random.randn(21,)
    if jax_ndarray:
        return jnp.array(X), jnp.array(y)
    return X, y


def get_dummy_image_data(jax_ndarray=True):
    X = onp.random.randn(21, 16, 16, 1)
    y = onp.random.randn(21,)
    if jax_ndarray:
        return jnp.array(X), jnp.array(y)
    return X, y


def get_dummy_vector_data(jax_ndarray=True):
    X, y = get_dummy_data(jax_ndarray)
    X = X[None].repeat(3, axis=0)
    y = y[None].repeat(3, axis=0)
    return X, y


class CustomConvNet(hk.Module):
    def __init__(self, embedim=2):
        super().__init__()
        self._embedim = embedim   

    def __call__(self, x):
        x = hk.Conv2D(32, 3)(x)
        x = jax.nn.relu(x)
        x = hk.MaxPool(2, 2, 'SAME')(x)
        x = hk.Conv2D(64, 3)(x)
        x = jax.nn.relu(x)
        x = hk.Flatten()(x)
        x = hk.Linear(self._embedim)(x)
        return x


@pytest.mark.parametrize("jax_ndarray", [True, False])
def test_single_fit(jax_ndarray):
    X, y = get_dummy_data(jax_ndarray)
    rng_key = get_keys()[0]
    m = viDKL(X.shape[-1])
    nn_params, kernel_params, losses = m.single_fit(
        rng_key, X, y, num_steps=100, step_size=0.05)
    assert isinstance(kernel_params, dict)
    assert isinstance(nn_params, dict)
    assert isinstance(losses, jnp.ndarray)


@pytest.mark.parametrize("jax_ndarray", [True, False])
def test_single_fit_custom_net(jax_ndarray):
    X, y = get_dummy_image_data(jax_ndarray)
    rng_key = get_keys()[0]
    m = viDKL(X.shape[1:], nn=CustomConvNet)
    nn_params, kernel_params, losses = m.single_fit(
        rng_key, X, y, num_steps=100, step_size=0.05)
    for i, val in enumerate(nn_params.values()):
        for k, v in val.items():
            if 'w' in k and i < 2:
                assert_equal(v.ndim, 4) # confirm that this is a 4-dim weights tensor of CNN


def test_get_mvn_posterior():
    rng_key = get_keys()[0]
    X, y = get_dummy_data()
    X_test, _ = get_dummy_data()
    net = hk.transform(lambda x: MLP()(x))
    nn_params = net.init(rng_key, X)
    kernel_params = {"k_length": jnp.array([1.0]),
                     "k_scale": jnp.array(1.0),
                     "noise": jnp.array(0.1)}
    m = viDKL(X.shape[-1])
    mean, cov = m.get_mvn_posterior(X, y, X_test, nn_params, kernel_params)
    assert isinstance(mean, jnp.ndarray)
    assert isinstance(cov, jnp.ndarray)
    assert_equal(mean.shape, (X_test.shape[0],))
    assert_equal(cov.shape, (X_test.shape[0], X_test.shape[0]))


def test_get_mvn_posterior_noiseless():
    rng_key = get_keys()[0]
    X, y = get_dummy_data()
    X_test, _ = get_dummy_data()
    net = hk.transform(lambda x: MLP()(x))
    nn_params = net.init(rng_key, X)
    kernel_params = {"k_length": jnp.array([1.0]),
                     "k_scale": jnp.array(1.0),
                     "noise": jnp.array(0.1)}
    m = viDKL(X.shape[-1])
    mean1, cov1 = m.get_mvn_posterior(X, y, X_test, nn_params, kernel_params, noiseless=False)
    mean1_, cov1_ = m.get_mvn_posterior(X, y, X_test, nn_params, kernel_params, noiseless=False)
    mean2, cov2 = m.get_mvn_posterior(X, y, X_test, nn_params, kernel_params, noiseless=True)
    assert_array_equal(mean1, mean1_)
    assert_array_equal(cov1, cov1_)
    assert_array_equal(mean1, mean2)
    assert onp.count_nonzero(cov1 - cov2) > 0


def test_fit_scalar_target():
    X, y = get_dummy_data()
    rng_key = get_keys()[0]
    m = viDKL(X.shape[-1])
    m.fit(rng_key, X, y, num_steps=100, step_size=0.05)
    for v in m.kernel_params.values():
        assert v.ndim < 2
    for val in m.nn_params.values():
        for v in val.values():
            assert v.ndim < 3

def test_fit_vector_target():
    X, y = get_dummy_vector_data()
    rng_key = get_keys()[0]
    m = viDKL(X.shape[-1])
    m.fit(rng_key, X, y, num_steps=100, step_size=0.05)
    for v in m.kernel_params.values():
        assert v.ndim > 0
        assert_equal(v.shape[0], 3)
    for val in m.nn_params.values():
        for v in val.values():
            assert v.ndim > 1
            assert_equal(v.shape[0], 3)


def test_predict_scalar():
    rng_key = get_keys()[0]
    X, y = get_dummy_data()
    X_test, _ = get_dummy_data()
    net = hk.transform(lambda x: MLP()(x))
    nn_params = net.init(rng_key, X)
    kernel_params = {"k_length": jnp.array([1.0]),
                     "k_scale": jnp.array(1.0),
                     "noise": jnp.array(0.1)}
    m = viDKL(X.shape[-1])
    m.X_train = X
    m.y_train = y
    m.nn_params = nn_params
    m.kernel_params = kernel_params
    mean, var = m.predict(rng_key, X_test)
    assert isinstance(mean, jnp.ndarray)
    assert isinstance(var, jnp.ndarray)
    assert_equal(mean.shape, (len(X_test),))
    assert_equal(var.shape, (len(X_test),))


def test_predict_vector():
    rng_key = get_keys()[0]
    X, y = get_dummy_vector_data()
    X_test, _ = get_dummy_vector_data()
    net = hk.transform(lambda x: MLP()(x))
    clone = lambda x: net.init(rng_key, x)
    nn_params = jax.vmap(clone)(X)
    kernel_params = {"k_length": jnp.array([[1.0], [1.0], [1.0]]),
                     "k_scale": jnp.array([1.0, 1.0, 1.0]),
                     "noise": jnp.array([0.1, 0.1, 0.1])}
    m = viDKL(X.shape[-1])
    m.X_train = X
    m.y_train = y
    m.nn_params = nn_params
    m.kernel_params = kernel_params
    mean, var = m.predict(rng_key, X_test)
    assert isinstance(mean, jnp.ndarray)
    assert isinstance(var, jnp.ndarray)
    assert_equal(mean.shape, X_test.shape[:-1])
    assert_equal(var.shape, X_test.shape[:-1])


def test_predict_in_batches_scalar():
    rng_key = get_keys()[0]
    X, y = get_dummy_data()
    X_test, _ = get_dummy_data()
    net = hk.transform(lambda x: MLP()(x))
    nn_params = net.init(rng_key, X)
    kernel_params = {"k_length": jnp.array([1.0]),
                     "k_scale": jnp.array(1.0),
                     "noise": jnp.array(0.1)}
    m = viDKL(X.shape[-1])
    m.X_train = X
    m.y_train = y
    m.nn_params = nn_params
    m.kernel_params = kernel_params
    mean, var = m.predict_in_batches(rng_key, X_test, batch_size=10)
    assert isinstance(mean, jnp.ndarray)
    assert isinstance(var, jnp.ndarray)
    assert_equal(mean.shape, (len(X_test),))
    assert_equal(var.shape, (len(X_test),))


def test_predict_in_batches_vector():
    rng_key = get_keys()[0]
    X, y = get_dummy_vector_data()
    X_test, _ = get_dummy_vector_data()
    net = hk.transform(lambda x: MLP()(x))
    clone = lambda x: net.init(rng_key, x)
    nn_params = jax.vmap(clone)(X)
    kernel_params = {"k_length": jnp.array([[1.0], [1.0], [1.0]]),
                     "k_scale": jnp.array([1.0, 1.0, 1.0]),
                     "noise": jnp.array([0.1, 0.1, 0.1])}
    m = viDKL(X.shape[-1])
    m.X_train = X
    m.y_train = y
    m.nn_params = nn_params
    m.kernel_params = kernel_params
    mean, var = m.predict_in_batches(rng_key, X_test, batch_size=10)
    assert isinstance(mean, jnp.ndarray)
    assert isinstance(var, jnp.ndarray)
    assert_equal(mean.shape, X_test.shape[:-1])
    assert_equal(var.shape, X_test.shape[:-1])


def test_fit_predict_scalar():
    rng_key = get_keys()[0]
    X, y = get_dummy_data()
    X_test, _ = get_dummy_data()
    m = viDKL(X.shape[-1])
    mean, var = m.fit_predict(
        rng_key, X, y, X_test, num_steps=100, step_size=0.05, batch_size=10)
    assert isinstance(mean, jnp.ndarray)
    assert isinstance(var, jnp.ndarray)
    assert_equal(mean.shape, (len(X_test),))
    assert_equal(var.shape, (len(X_test),))


def test_fit_predict_vector():
    rng_key = get_keys()[0]
    X, y = get_dummy_vector_data()
    X_test, _ = get_dummy_vector_data()
    m = viDKL(X.shape[-1])
    mean, var = m.fit_predict(
        rng_key, X, y, X_test, num_steps=100, step_size=0.05, batch_size=10)
    assert isinstance(mean, jnp.ndarray)
    assert isinstance(var, jnp.ndarray)
    assert_equal(mean.shape, X_test.shape[:-1])
    assert_equal(var.shape, X_test.shape[:-1])


def test_fit_predict_scalar_ensemble():
    rng_key = get_keys()[0]
    X, y = get_dummy_data()
    X_test, _ = get_dummy_data()
    m = viDKL(X.shape[-1])
    mean, var = m.fit_predict(
        rng_key, X, y, X_test, n_models=4,
        num_steps=100, step_size=0.05, batch_size=10)
    assert isinstance(mean, jnp.ndarray)
    assert isinstance(var, jnp.ndarray)
    assert_equal(mean.shape, (4, len(X_test),))
    assert_equal(var.shape, (4, len(X_test),))


def test_fit_predict_vector_ensemble():
    rng_key = get_keys()[0]
    X, y = get_dummy_vector_data()
    X_test, _ = get_dummy_vector_data()
    m = viDKL(X.shape[-1])
    mean, var = m.fit_predict(
        rng_key, X, y, X_test, n_models=2,
        num_steps=100, step_size=0.05, batch_size=10)
    assert isinstance(mean, jnp.ndarray)
    assert isinstance(var, jnp.ndarray)
    assert_equal(mean.shape, (2, *X_test.shape[:-1]))
    assert_equal(var.shape, (2, *X_test.shape[:-1]))


def test_fit_predict_scalar_ensemble_custom_net():
    rng_key = get_keys()[0]
    X, y = get_dummy_image_data()
    X_test, _ = get_dummy_image_data()
    m = viDKL(X.shape[1:], nn=CustomConvNet)
    mean, var = m.fit_predict(
        rng_key, X, y, X_test, n_models=2,
        num_steps=100, step_size=0.05, batch_size=10)
    assert isinstance(mean, jnp.ndarray)
    assert isinstance(var, jnp.ndarray)
    assert_equal(mean.shape, (2, len(X_test),))
    assert_equal(var.shape, (2, len(X_test),))



"""
utils.py
========

Utility functions

Created by Maxim Ziatdinov (email: maxim.ziatdinov@ai4microscopy.com)
"""

from typing import Union, Dict, Type

import jax
import jax.numpy as jnp
import numpy as onp

import numpyro


def enable_x64():
    """Use double (x64) precision for jax arrays"""
    jax.config.update("jax_enable_x64", True)


def get_keys(seed: int = 0):
    """
    Simple wrapper for jax.random.split to get
    rng keys for model inference and prediction
    """
    rng_key_1, rng_key_2 = jax.random.split(jax.random.PRNGKey(seed))
    return rng_key_1, rng_key_2


def split_in_batches(X_new: Union[onp.ndarray, jnp.ndarray],
                     batch_size: int = 100, dim: int = 0):
    """
    Splits array into batches along the first or second dimensions
    """
    if dim not in [0, 1]:
        raise NotImplementedError("'dim' must be equal to 0 or 1")
    num_batches = jax.numpy.floor_divide(X_new.shape[dim], batch_size)
    X_split = []
    for i in range(num_batches):
        if dim == 0:
            X_i = X_new[i*batch_size:(i+1)*batch_size]
        else:
            X_i = X_new[:, i*batch_size:(i+1)*batch_size]
        X_split.append(X_i)
    X_i = X_new[(i+1)*batch_size:] if dim == 0 else X_new[:, (i+1)*batch_size:]
    if X_i.shape[dim] > 0:
        X_split.append(X_i)
    return X_split


def get_haiku_dict(kernel_params: Dict[str, jnp.ndarray]) -> Dict[str, Dict[str, jnp.ndarray]]:
    """
    Extracts weights and biases from viDKL dictionary into a separate
    dictionary compatible with haiku's .apply() method
    """
    all_weights = {}
    all_biases = {}
    for key, val in kernel_params.items():
        if key.startswith('feature_extractor'):
            name_split = key.split('/')
            name_new = name_split[1] + '/' + name_split[2][:-2]
            if name_split[2][-1] == 'b':
                all_biases[name_new] = val
            else:
                all_weights[name_new] = val
    nn_params = {}
    for (k, v1), (_, v2) in zip(all_weights.items(), all_biases.items()):
        nn_params[k] = {"w": v1, "b": v2}
    return nn_params


def dviz(d: Type[numpyro.distributions.Distribution], samples: int = 1000) -> None:
    """
    Utility function for visualizing numpyro distributions

    Args:
        d: numpyro distribution; e.g. numpyro.distributions.Gamma(2, 2)
        samples: number of samples
    """
    try:
        import seaborn as sns  # noqa: F401
    except ImportError as e:
        raise ImportError(
            "You need to install `seaborn` to be able to use this feature. "
            "It can be installed with `pip install seaborn`."
        ) from e
    import matplotlib.pyplot as plt

    with numpyro.handlers.seed(rng_seed=0):
        samples = d.sample(jax.random.PRNGKey(0), sample_shape=(samples,))
    plt.figure(dpi=100)
    sns.histplot(samples, kde=True, fill=False)
    plt.show()


def preprocess_sparse_image(sparse_image):
    """
    Creates GP inputs from sparse image data where missing values are represented by zeros.
    If your actual data contains zeros, you will need to (re-)normalize it.
    Otherwise, those elements will be interpreted as missng values. The function returns
    two arrays of the shapes (N, D) and (N,) that are used as training inputs and targets in GP
    and an array of full indices of the shape (N_full, D) for reconstructing the full image. D is
    the image dimensionality (D=2 for a 2D image)
    """
    # Find non-zero element indices
    non_zero_indices = onp.nonzero(sparse_image)
    # Create the GP input using the indices
    gp_input = onp.column_stack(non_zero_indices)
    # Extract non-zero values (targets) from the sparse image
    targets = sparse_image[non_zero_indices]
    # Generate indices for the entire image
    full_indices = onp.array(onp.meshgrid(*[onp.arange(dim) for dim in sparse_image.shape])).T.reshape(-1, sparse_image.ndim)
    return gp_input, targets, full_indices



"""
hypo.py
========

Utility functions for hypothesis learning based on arXiv:2112.06649

Created by Maxim Ziatdinov (email: maxim.ziatdinov@ai4microscopy.com)
"""

from typing import Callable, Dict, Optional, Union

import jax.numpy as jnp
import numpy as np
import numpyro

from .gp import ExactGP
from .spm import sPM
from .utils import get_keys


def step(model: Callable[[jnp.ndarray, Dict[str, jnp.ndarray]], jnp.ndarray],
         model_prior: Callable[[], Dict[str, jnp.ndarray]],
         X_measured: jnp.ndarray, y_measured: jnp.ndarray,
         X_unmeasured: Optional[jnp.ndarray] = None,
         gp_wrap: Optional[bool] = False,
         noise_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
         gp_kernel: str = 'Matern',
         gp_kernel_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
         gp_input_dim: Optional[int] = 1,
         num_warmup: Optional[int] = 2000, num_samples: Optional[int] = 2000,
         num_chains: Optional[int] = 1,
         num_restarts: Optional[int] = 1,
         print_summary: Optional[bool] = True):
    """
    Compute model posterior and use it to derive predictive uncertainty

    Args:
        model:
            Parametric model in jax.numpy
        model_prior:
            Prior over model parameters using numpyro.distributions
        X_measured:
            Measured points
        y_measured:
            Measured values
        X_unmeasured:
            Unmeasured points
        gp_wrap:
            Wrap probabilistic model into a Gaussian process (Default: False)
        noise_prior:
            Custom prior for observation noise. Defaults to LogNormal(0,1)
        gp_kernel:
            Gaussian process kernel (if gp_wrap is True). Defaults to Matern
        gp_kernel_prior:
            Custom priors over kernel hyperparameters. Defaults to LogNormal(0,1)
        gp_input_dim:
            Number of lenghscale dimensions in GP kernel.
            Equals to number of input dimensions or 1 (default)
        num_warmup:
            Number of warmup steps for HMC. Defaults to 2000
        num_samples:
            Number of HMC samples. Defaults to 2000
        num_chains:
            Number of HMC chains. Defaults to 2000
        num_restarts:
            Number of restarts if r_hat values are not acceptable (>1.1).
            Defaults to 1
        print_summary:
            Verbose parameter

    Returns:
        Predictive uncertainty and trained model object
    """
    verbose = print_summary
    sgr = numpyro.diagnostics.split_gelman_rubin
    for i in range(num_restarts):
        rng_key, rng_key_predict = get_keys(i)
        # Get/update model posterior
        if gp_wrap:  # wrap model into a gaussian process (gives more flexibility)
            model_ = ExactGP(
                gp_input_dim, gp_kernel, model,
                gp_kernel_prior, model_prior, noise_prior)
            model_.fit(
                rng_key, X_measured, y_measured, num_warmup,
                num_samples, num_chains, print_summary=verbose)
        else:  # use a standalone model
            model_ = sPM(model, model_prior, noise_prior)
            model_.fit(
                rng_key, X_measured, y_measured, num_warmup,
                num_samples, num_chains, print_summary=verbose)
        rhats = [sgr(v).item() for (k,v) in model_.get_samples(1).items() if k != 'mu']
        if max(rhats) < 1.1:
            break
    # compute predictive uncertainty for the unmeasured part of the parameter space
    obj = 0
    if X_unmeasured is not None:
        mean, samples = model_.predict(rng_key, X_unmeasured)
        obj = samples.squeeze().var(0)
    return obj, model_


def sample_next(rewards: Union[np.array, jnp.array],
                method: Optional[str] = 'softmax',
                temperature: Optional[float] = 1.0,
                eps: Optional[float] = 0.4) -> int:
    """
    Sample model or input channel based on softmax or epsilon-greedy policy

    Args:
        rewards:
            Array of shape (N,) with running rewards
        method:
            Selection policy, choose between 'softmax' and 'epsilon-greedy'
        temperature:
            Optional temperature parameter for softmax selection policy
        eps:
            Optional epsilon parameter for epsilon-greedy policy

    Returns:
        The index of model or input channel to sample next
    """
    if method not in ['softmax', 'eps-greedy']:
        raise NotImplementedError(
            "The currently implemented samplong methods are 'softmax' and 'eps-greedy'")
    if rewards.ndim != 1:
        raise AttributeError("Pass rewards as 1-dimensional array")
    if method == 'softmax':
        idx = softmax(rewards, temperature)
    else:
        idx = eps_greedy(rewards, eps)
    return idx


def softmax(logits: Union[np.array, jnp.array],
            temperature: Optional[float] = 1.0) -> int:
    """
    Softmax selection policy. Based on
    Zai, A., Brown, B. (2020). Deep reinforcement learning in action. Manning Publications
    """
    probs = np.exp(logits / temperature) / np.sum(np.exp(logits / temperature))
    x = np.arange(len(logits))
    idx = np.random.choice(x, p=probs)
    return idx


def eps_greedy(rewards: Union[np.array, jnp.array],
               eps: Optional[float] = 0.4) -> int:
    """
    Epsilon-greedy selection policy. Based on
    Zai, A., Brown, B. (2020). Deep reinforcement learning in action. Manning Publications
    """
    if np.random.random() > eps:
        idx = rewards.argmax()
    else:
        idx = np.random.randint(len(rewards))
    return idx


def update_record(record: np.array, action: int, r: Union[int, float]) -> np.array:
    """
    Update the reward record. Based on
    Zai, A., Brown, B. (2020). Deep reinforcement learning in action. Manning Publications
    """
    new_r = (record[action, 0] * record[action, 1] + r) / (record[action, 0] + 1)
    record[action, 0] += 1
    record[action, 1] = new_r
    return record



from . import utils, kernels, acquisition
from .gp import ExactGP
from .vgp import vExactGP
from .dkl import DKL
from .vidkl import viDKL
from .vigp import viGP
from .spm import sPM
from .hypo import sample_next

from .__version__ import version as __version__

__all__ = ["utils", "kernels", "acquisition", "ExactGP", "vExactGP", "DKL",
           "viDKL", "viGP", "sPM", "sample_next", "__version__"]
